"""
This module implements node objects which represent operations on tabular 
datasets and can be combined together to assemble SQL queries. In the `render` 
step, FunSQL compiles the assembled node tree into a SQL query. 

NOTE [DEV]:
1. Writing the __init__ method gets repetitive, but dataclasses don't help with star-arg inputs. 
Also, we can't hint at what input types we allow, annotations only show what it gets coerced to. 
2. Can abstract out the `rebase` and `pretty print` method but there are a lot of edge cases, so 
lets just keep it simple. The compiler is the beast part of the code. 
"""

from functools import partial
from typing import Any, Optional, Union, Callable

from .common import Symbol, S, LITERAL_TYPES, register_union_type
from .sqlcontext import SQLTable, ValuesTable
from .nodes import *
from .clausedefs import (
    Frame,
    FrameMode,
    FrameEdgeSide,
    FrameEdge,
    FrameExclude,
    ValueOrder,
    NullsOrder,
)
from .prettier import (
    Doc,
    call_expr,
    assg_expr,
    list_expr,
    pipe_expr,
    to_doc,
    QuoteContext,
)

__all__ = [
    "Agg",
    "Append",
    "As",
    "Bind",
    "Define",
    "From",
    "Fun",
    "Get",
    "Group",
    "Iterate",
    "Join",
    "Limit",
    "Lit",
    "Order",
    "Partition",
    "Select",
    "Sort",
    "Var",
    "Where",
    "With",
    "Asc",
    "Desc",
    "ValuesTable",
    "aka",
    "eq",
    "LeftJoin",
    "F",
]


# -----------------------------------------------------------
# utility functions local to this module
# -----------------------------------------------------------

# These are internal
RESERVED_WORDS = {
    "over",
    "args",
    "filter_",
    "distinct",
}


class ExtendAttrs(type):
    """
    A SQLNode using this as a metaclass can support fluent syntax like
    `Fun.count(*args)` instead of `Fun("count", *args)`. Used by the `Fun`
    and `Agg` classes.

    NOTE: This is a frequence source of bugs, since the user provided key value
    might clash with an attribute of the class. Look out for it probably?
    """

    def __getattr__(cls, key: str) -> Callable:
        if key in RESERVED_WORDS:
            raise AttributeError(
                f"{key} is a reserved word, please use the call expression syntax instead; `Get(...)` or `Agg(...)`"
            )
        return partial(cls, key)


class ExtendAttrsFull(type):
    """
    Same as ExtendAttrs but actually initializes the class. Used by the
    `Get` and `Var` classes.
    """

    def __getattr__(cls, key: str) -> Any:
        if key in RESERVED_WORDS:
            raise AttributeError(
                f"{key} is a reserved word, please use the call expression syntax instead; `Get(...)` or `Agg(...)`"
            )
        return cls(key)


def _rebase_node(curr: Optional[SQLNode], pre: SQLNode) -> SQLNode:
    """shorthand for a common pattern"""
    if curr is None:
        return pre
    else:
        return curr.rebase(pre)


def _cast_to_node(node: Union[SQLNode, Any]) -> SQLNode:
    """Convert a value to a SQLNode"""
    if isinstance(node, SQLNode):
        return node
    else:
        return Lit(node)


def _cast_to_node_skip_none(node: Union[SQLNode, Any]) -> Optional[SQLNode]:
    """Convert a value to a SQLNode, skipping None"""
    if node is None:
        return None
    else:
        return _cast_to_node(node)


# -----------------------------------------------------------
# Node definitions to export
# -----------------------------------------------------------


class Agg(SQLNode, metaclass=ExtendAttrs):
    """
    `Agg` is used to create an aggregate expression over a column. It should be applied
    only within the output of nodes of type:
    * `Group`: gets translated to a column aggregate for the `GROUP BY` operation
    * `Partition`: gets translated to a `WINDOW` function.

    Specific database engines may support additional aggregate functions besides the ones
    that are common - count, sum, max, min, etc. FunSQL is unaware of what is supported, and
    renders all of them as regular function expressions.

    Args:
        name: name of the aggregate function
        *args: arguments to the aggregate function
        distinct: whether to use only distinct values from the column for aggregation
        filter_: rows to filter out from the dataset before aggregation

    Examples:
    ```python
    # To fetch the number of students in each class
    >>> q = From(Students) >> Group(Get.class_id) >> Select(Get.class_id, Agg.count())

    # Compute average playing time of players in the Arsenal first team
    >>> q = (
            From(Players)
            >> Where(eq(Get.team_id, "Arsenal"))
            >> Group(Get.player_id)
            >> Select(Get.player_id, Agg.avg(Get.playing_time)
        )

    # Add a column indicating the rank of each building by height in its city
    >>> q = (
            From(Buildings)
            >> Partition(Get.city_id, order_by=[Get.height])
            >> Select(Get.city_id, Get.building_id, Get.height, Agg.row_number()
        )
    """

    _name: Symbol
    args: list[SQLNode]
    distinct: bool
    filter_: Optional[SQLNode]
    over: Optional[SQLNode]

    def __init__(
        self,
        name: Union[str, Symbol],
        *args: NODE_MATERIAL,
        distinct: bool = False,
        filter_: Optional[NODE_MATERIAL] = None,
        over: Optional[SQLNode] = None,
    ) -> None:
        super().__init__()
        self._name = S(name)
        self.args = [_cast_to_node(arg) for arg in args]
        self.distinct = distinct
        self.filter_ = _cast_to_node_skip_none(filter_)
        self.over = _cast_to_node_skip_none(over)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(
            self._name,
            *self.args,
            distinct=self.distinct,
            filter_=self.filter_,
            over=_rebase_node(self.over, pre),
        )

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "Agg"
        args = []

        _name_str = str(self._name)
        if _name_str.isalpha():
            name = f"Agg.{_name_str}"  # Agg.Count
        else:
            name = f'Agg."{_name_str}"'  # edge cases

        if self.distinct:
            args.append(assg_expr("distinct", to_doc(self.distinct)))
        for arg in self.args:
            args.append(to_doc(arg, ctx))
        if self.filter_ is not None:
            args.append(assg_expr("filter", to_doc(self.filter_, ctx)))
        if self.over is not None:
            args.append(assg_expr("over", to_doc(self.over, ctx)))

        return call_expr(name, args)


class Append(TabularNode):
    """
    `Append` concatenates input datasets, analogous to a `UNION ALL` operation in SQL.
    Only the columns present in all the input datasets are kept.

    Args:
        *args: input datasets to take a union over

    Examples:
    ```python

    # Collect the birthdays of everyone in the School
    >>> q = (
        From(Students)
        >> Select(Get.date_of_birth)
        >> Append(
            From(Teachers) >> Select(Get.date_of_birth),
            From(Staff) >> Select(Get.date_of_birth)
        )
    )
    ```
    """

    args: list[SQLNode]
    over: Optional[SQLNode]

    def __init__(self, *args: NODE_MATERIAL, over: Optional[SQLNode] = None) -> None:
        super().__init__()
        self.args = [_cast_to_node(arg) for arg in args]
        self.over = _cast_to_node_skip_none(over)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(*self.args, over=_rebase_node(self.over, pre))

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "Append"
        args = []

        if len(self.args) == 0:
            args.append(assg_expr("args", to_doc("[]")))
        else:
            for arg in self.args:
                args.append(to_doc(arg, ctx))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


class As(SQLNode):
    """
    `As` node is similar but not quite like the `AS` clause in SQL. When preceded by a:
        * scalar expression - it specifies the name of the column to refer to it
        * tabular node - it nests the namespace for the subquery behind the assigned name

    When used in a tabular context, the alias provided by `As` is used by FunSQL to identify
    column references correctly. It might not end up in the rendered SQL query if not necessary.

    Args:
        name: name assigned to the preceding node
        over: scalar/subquery being assigned the name

    Examples:
    ```python
    # rename the id column when reading from a table
    >>> q = From(Students) >> Select(Get.id >> As("student_id"))

    # nest the location table behind the alias "location" to disambiguate between the `id` column
    # of the `Students` table and the `id` column of the `Location` table.
    >>> q = (
            From(Students)
            >> Join(Location >> As("location"), eq(Get.location.id, Get.city_id))
            >> Select(Get.id, Get.name, Get.location.name)
        )
    ```
    """

    name: Symbol
    over: Optional[SQLNode]

    def __init__(
        self, name: Union[str, Symbol], over: Optional[NODE_MATERIAL] = None
    ) -> None:
        super().__init__()
        self.name = S(name)
        self.over = _cast_to_node_skip_none(over)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(name=self.name, over=_rebase_node(self.over, pre))

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "As"
        args: list["Doc"] = [str(self.name)]

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


class Bind(SQLNode):
    """
    `Bind` evaulates the parent subquery with parameters; that is maps `Var` variables in
    the parent node to the supplied values.

    Bind is especially useful for correlated query cases. When a subquery with `Bind` applied is
    used as a scalar expression, it gets rendered as a correlated subquery. While when `Bind` is
    applied to the `joinee` branch of a Join node, it gets rendered as a lateral join.

    Args:
        *args: nodes in the form, `value >> As(variable_name)`

    Examples:
    ```python
    # fetch the number of times a person visited their healthcare provider
    >>> q1 = From(Visits) >> Where(eq(Get.patient_id, Var.PERSON_ID)) >> Group() >> Select(Agg.count())
    # This gives us a `prepared statement` in SQL, which can be executed with different values for the `PERSON_ID` parameter.

    # Alternatively, `Bind` lets us supply the parameter to create the full query directly.
    >>> q = q1 >> Bind(Var.PERSON_ID, 1000)

    # By binding column references instead of explicity values, we can create correlated queries.
    # For example, to fetch patients with at least one visit,
    >>> visited = lambda x: From(Visits) >> Where(eq(Get.patient_id, Var.PERSON_ID)) >> Bind(x >> As("PERSON_ID"))
    >>> q = From(Person) >> Where(Fun.exists(visits(Get.person_id))

    # To fetch patients along with their last visit date, we use `Bind` in a tabular context.
    >>> q = (
            From(person)
            >> Join(From(Visits)
                    >> Where(eq(Get.patient_id, Var.PERSON_ID))
                    >> Order(Get.visit_date >> Desc())
                    >> Limit(1)
                    >> Bind(Get.person_id >> As("PERSON_ID"))
                    >> As("last_visit"),
                    on=True,
                    left=True
                )
            >> Select(Get.person_id, Get.name, Get.last_visit.visit_date)
        )
    ```
    """

    args: list[SQLNode]
    over: Optional[SQLNode]
    label_map: dict[Symbol, int]

    def __init__(self, *args: NODE_MATERIAL, over: Optional[SQLNode] = None) -> None:
        super().__init__()
        self.args = [_cast_to_node(arg) for arg in args]
        self.over = _cast_to_node_skip_none(over)
        self.label_map = populate_label_map(self, self.args)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(*self.args, over=_rebase_node(self.over, pre))

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "Bind"
        args = []
        if len(self.args) == 0:
            args.append(assg_expr("args", to_doc("[]")))
        else:
            for arg in self.args:
                args.append(to_doc(arg, ctx))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


class Define(TabularNode):
    """
    Define node creates or replaces columns in the parent subquery.

    Args:
        *args: nodes in the form, `column reference >> As(column_name)`

    Examples:
    ```python
    # Define a new column listing the age of each person
    >>> q = From(Person) >> Define(Fun("-", Fun.now(), Get.date_of_birth) >> As("age")) >> Select(Get.name, Get.age)

    # Compute the discounted price, and after adding taxes. Define nodes make query writing modular.
    >>> q = (
            From(Cart)
            >> Join(From(Product) >> As("product"), on=eq(Get.product_id, Get.product.id), left=True)
            >> Define(Fun("-", 1, Get.discount) >> As("discount_factor"))
            >> Define(Fun("*", Get.product.price, Get.discount_factor) >> As("discounted_price"))
            >> Define(Fun("*", Get.discounted_price, Get.product.tax_rate) >> As("tax_amount"))
            >> Define(Fun("+", Get.discounted_price, Get.tax_amount) >> As("total_price"))
        )
    ```
    """

    args: list[SQLNode]
    over: Optional[SQLNode]
    label_map: dict[Symbol, int]

    def __init__(self, *args: NODE_MATERIAL, over: Optional[SQLNode] = None) -> None:
        super().__init__()
        self.args = [_cast_to_node(arg) for arg in args]
        self.over = _cast_to_node_skip_none(over)
        self.label_map = populate_label_map(self, self.args)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(*self.args, over=_rebase_node(self.over, pre))

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "Define"
        args = [to_doc(arg, ctx) for arg in self.args]

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


class From(TabularNode):
    """
    `From` refers to the contents of a database table; present at the root of query objects.
    The source of a `From` node can be of the type:
    * `SQLTable`: corresponds to a table in the database. Downstream nodes can refer to its columns, and build expressions from them.
    * `Symbol`: refers to a table provided at render time, or a CTE subquery constructed using a `With` node.
    * `ValuesTable`: refers to a list of tuples of values, gets rendered as a VALUES clause in the final SQL.
    * None: represents a subquery with one row and no columns.

    Args:
        source: table, symbol or a Values table.

    Examples:

    # To select from the table with user address details
    >>> Addresses = SQLTable(S.addresses, columns=["id", "user_id", "address", "city"])
    >>> q = From(Addresses) >> Select(Get.user_id, Get.city)

    # To filter out the discounted products in the catalog
    >>> q = (
            From(Products)
            >> Where(Fun.IN(Get.id, From(S.discounted_items) >> Select(Get.product_id)))
            >> With(From(Discounts) >> Where(Fun(">", Get.discount, 0)) >> As("discounted_items"))
        )
    """

    source: Union[Symbol, SQLTable, ValuesTable, None]

    def __init__(
        self, source: Union[Symbol, SQLTable, ValuesTable, None] = None
    ) -> None:
        super().__init__()
        if isinstance(source, (SQLTable, ValuesTable)) or source is None:
            self.source = source
        else:
            self.source = S(source)

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "From"
        args = []

        if isinstance(self.source, SQLTable):
            alias = ctx.vars_.get(self.source, None)
            if alias is None:
                args.append(to_doc(self.source, QuoteContext(limit=True)))
            else:
                args.append(str(alias))
        elif isinstance(self.source, Symbol):
            args.append(str(self.source))
        elif isinstance(self.source, ValuesTable):
            if ctx.limit:
                args.append("...")
            else:
                for col in self.source.columns:
                    args.append(assg_expr(str(col), "[...]"))

        return call_expr(name, args)


class Fun(SQLNode, metaclass=ExtendAttrs):
    """
    `Fun` is used to apply a SQL function or operator.

    FunSQL is unaware of which functions/operators are actually supported by each
    database engine. By default, `Fun` node is translated to the function call syntax
    by default, if the `name` argument is a valid python identifier, else it is rendered
    as an operator. For the commonly used function names below, FunSQL specifies the
    rendering explicitly.

    |---------------------------------- |---------------------------------------|
    | `Fun` node                        | SQL syntax                            |
    |---------------------------------- |---------------------------------------|
    | `Fun("=", x, y)`                  | `x = y`                               |
    | `Fun."!="(x, y)`                  | `x <> y`                              |
    | `Fun.and(p₁, p₂, …)`              | `p₁ AND p₂ AND …`                     |
    | `Fun.between(x, y, z)`            | `x BETWEEN y AND z`                   |
    | `Fun.case(p, x, …)`               | `CASE WHEN p THEN x … END`            |
    | `Fun.cast(x, {TYPE})`             | `CAST(x AS {TYPE})`                   |
    | `Fun.current_date()`              | `CURRENT_DATE`                        |
    | `Fun.current_timestamp()`         | `CURRENT_TIMESTAMP`                   |
    | `Fun.exists(q)`                   | `EXISTS q`                            |
    | `Fun.extract("FIELD", x)`         | `EXTRACT(FIELD FROM x)`               |
    | `Fun.in(x, q)`                    | `x IN q`                              |
    | `Fun.in(x, y₁, y₂, …)`            | `x IN (y₁, y₂, …)`                    |
    | `Fun("is not null", x)`           | `x IS NOT NULL`                       |
    | `Fun("is null", x)`               | `x IS NULL`                           |
    | `Fun.like(x, y)`                  | `x LIKE y`                            |
    | `Fun.not(p)`                      | `NOT p`                               |
    | `Fun("not between", x, y, z)`     | `x NOT BETWEEN y AND z`               |
    | `Fun("not in", x, q)`             | `x NOT IN q`                          |
    | `Fun("not in", x, y₁, y₂, …)`     | `x NOT IN (y₁, y₂, …)`                |
    | `Fun.or(p₁, p₂, …)`               | `p₁ OR p₂ OR …`                       |
    |---------------------------------- |---------------------------------------|


    Args:
        name: name of the function or operator.
        *args: arguments to the function or operator.

    Examples:
    ```python
    # filter out users named "John"
    >>> q = From(Users) >> Where(Fun.like(Get.name, "%John%"))

    # specify the Join condition for two tables
    >>> q = From(Users) >> Join(Orders >> As(S.orders), Fun("=", Get.id, Get.orders.user_id))
    ```
    """

    _name: Symbol
    args: list[SQLNode]

    def __init__(self, name: Union[str, Symbol], *args: NODE_MATERIAL) -> None:
        super().__init__()
        self._name = S(name)
        self.args = [_cast_to_node(arg) for arg in args]

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "Fun"
        args = []

        _name_str = str(self._name)
        if _name_str.isalpha():
            name = f"Fun.{_name_str}"  # Fun.exists
        else:
            name = f'Fun."{_name_str}"'  # Fun.">"

        for arg in self.args:
            args.append(to_doc(arg, ctx))
        return call_expr(name, args)


class Get(SQLNode, metaclass=ExtendAttrsFull):
    """
    `Get` node creates reference to a column in the input dataset/subquery.

    To disambiguate between columns with the same name (like the join output of
    two tables), one of the tables is wrapped in an `As` node. To access wrapped
    columns, the chained syntax for `Get is used: `Get.A.B.C` refers to a column
    wrapped as `... >> As(S.B) >> As(S.A)`. The chained Get node, can also be created
    as: `Get.A >> Get.B >> Get.C`.

    Args:
        name: name of the column reference
        over: name wrapping the dataset which has the column

    Examples:
    ```python
    # get the name of the user
    >>> q = From(Users) >> Select(Get.name)

    # get the name of the user and the num of orders they made
    >>> q = (
            From(Users)
            >> Join(Orders >> As(S.orders), on=eq(Get.id, Get.orders.user_id))
            >> Select(Get.name, Get.orders.count())
        )

    # dereferencing a wrapped column
    >>> q = From(Users) >> As("A") >> As("B") >> As("C") >> Select(Get.C.B.A.name)
    ```
    """

    _name: Symbol
    over: Optional[SQLNode]

    def __init__(
        self, name: Union[str, Symbol], over: Optional[SQLNode] = None
    ) -> None:
        super().__init__()
        self._name = S(name)
        self.over = _cast_to_node_skip_none(over)

    def __getattr__(self, name: str) -> "Get":
        return self.__class__(name=S(name), over=self)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(name=self._name, over=_rebase_node(self.over, pre))

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        path = [self._name]
        over = self.over

        while over is not None and isinstance(over, Get) and over not in ctx.vars_:
            path.append(over._name)
            over = over.over
        if over is not None and over in ctx.vars_:
            path.append(ctx.vars_[over])
            over = None
        else:
            path.append(S("Get"))

        ex = ".".join(str(p) for p in reversed(path))
        if over is not None:
            ex = pipe_expr(to_doc(over, ctx), ex)
        return ex


class Group(TabularNode):
    """
    The `Group` node is akin to the SQL `GROUP BY` clause. It partitions the input dataset
    by the grouping key (set of columns specified). and creates a new dataset with exactly
    one row for each distinct value of the grouping key.
    More columns can be added to the grouped  dataset by creating aggregations using the
    `Agg` node. An aggregation computes a unique value for each group, summarizing it as specified.
    If no aggregations are defined, `Group` renders just the dataset with the grouping keys.

    NOTE: `Group` requires unique names for all the columns in the grouping key. So, the downstream
    nodes can refer to these columns directly, without needing a hierarchical `Get` node.

    args:
        *by: column references to group by

    Examples:
    ```python
    # count the total numbrt of users
    >>> q = From(Users) >> Group() >> Select(Agg.count())

    # count the total number of users by city
    >>> q = From(Users) >> Group(Get.city) >> Select(Get.city, Agg.count())

    # all the distinct cities users come from
    >>> q = From(Users) >> Group(Get.city)
    ```
    """

    by: list[SQLNode]
    over: Optional[SQLNode]
    label_map: dict[Symbol, int]

    def __init__(self, *by: NODE_MATERIAL, over: Optional[SQLNode] = None) -> None:
        super().__init__()
        self.by = [_cast_to_node(arg) for arg in by]
        self.over = _cast_to_node_skip_none(over)
        self.label_map = populate_label_map(self, self.by)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(*self.by, over=_rebase_node(self.over, pre))

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "Group"
        args = [to_doc(arg, ctx) for arg in self.by]

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


class Iterate(TabularNode):
    """
    Iterate is used to generate a recursive CTE. The Postgresql [docs](https://www.postgresql.org/docs/current/queries-with.html#QUERIES-WITH-RECURSIVE)
    do a good job explaining it. FunSQL translates the Iterate node roughly as:

    ```sql
    WITH RECURSIVE `iterator` AS (
        SELECT ...
        FROM `over`
        UNION ALL
        SELECT ...
        FROM `iterator`
    )
    SELECT ...
    FROM `iterator`
    ```

    Args:
        iterator: query that is executed repeatedly
        over: query that provides the initial value

    Examples:
    ```python
    # calculating the factorial function
    >>> q = (
            Define(1 >> As(S.n), 1 >> As(S.fact))
            >> Iterate(
                From(S.factorial)
                >> Define(Fun("+", Get.n, 1) >> As(S.n), Fun("*", Get.n, Get.fact) >> As(S.fact))
                >> Where(Fun("<", Get.n, 10))
                >> As(S.factorial)
            )
        )
    ```
    """

    iterator: SQLNode
    over: Optional[SQLNode]

    def __init__(self, iterator: SQLNode, over: Optional[SQLNode] = None) -> None:
        super().__init__()
        self.iterator = _cast_to_node(iterator)
        self.over = _cast_to_node_skip_none(over)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(iterator=self.iterator, over=_rebase_node(self.over, pre))

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "Iterate"
        args = [to_doc(self.iterator, ctx)]

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


class Join(TabularNode):
    """
    Join represents the output of a SQL join operation between two tables. The type of join is
    deduced from the input args:
    * INNER JOIN - default
    * LEFT JOIN - `left` is True
    * RIGHT JOIN - `right` is True
    * FULL JOIN - `left` is True, and `right` is True
    * CROSS JOIN - `on` is set to True

    FunSQL renders a lateral join when a variable in the `joinee` branch is binded to a column
    reference in the `over` subquery. Refer to the `Bind` node for an example.

    If the two sides of the join might have the same column names, disambiguate them by wrapping one
    of the tables using the `As` node.

    Args:
        joinee: table on right side of the join
        on: join condition
        left: if it is a left join
        right: if it is a right join
        skip: if True, the Join operation isn't rendered when right side table is not referenced downstream
        over: table on left side of the join

    NOTE: `skip` set to True leaves out the right side of the join if it doesn't contribute any column
    refs downstream. It isn't really desirable though, like in case of an inner join, the right table still
    affects the output through the join condition, so omitting it yields a different query.

    Examples:
    ```python
    # fetch the maximum order value for each user
    >>> q = (
            From(User)
            >> Join(From(Orders) >> As(S.orders), on=Fun("=", Get.id, Get.orders.user_id))
            >> Select(Get.id, Agg.max(Get.orders.value))
        )
    ```
    """

    joinee: SQLNode
    on: SQLNode
    left: bool
    right: bool
    skip: bool
    over: Optional[SQLNode]

    def __init__(
        self,
        joinee: SQLNode,
        on: NODE_MATERIAL,
        left: bool = False,
        right: bool = False,
        skip: bool = False,
        over: Optional[SQLNode] = None,
    ) -> None:
        super().__init__()
        self.joinee = _cast_to_node(joinee)
        self.on = _cast_to_node(on)
        self.left = left
        self.right = right
        self.skip = skip
        self.over = _cast_to_node_skip_none(over)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(
            joinee=self.joinee,
            on=self.on,
            left=self.left,
            right=self.right,
            skip=self.skip,
            over=_rebase_node(self.over, pre),
        )

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "Join"
        args = []

        if ctx.limit:
            args.append("...")
        else:
            args.append(to_doc(self.joinee, ctx))
            args.append(to_doc(self.on, ctx))
            if self.left:
                args.append(assg_expr("left", "True"))
            if self.right:
                args.append(assg_expr("right", "True"))
            if self.skip:
                args.append(assg_expr("skip", "True"))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


class Limit(TabularNode):
    """
    `Limit` limits the number of rows returned by a subquery. It is generally used in
    conjunction with the `Order` node, since the query output might not be deterministic
    otherwise.

    Args:
        limit: number of rows to return
        offset: number of rows to skip from the first position

    Examples:
    ```python
    # fetch any 10 user records from the database
    >>> q = From(Users) >> Select(Get.id, Get.name) >> Limit(10)

    # fetch the first 100 users by the date they joined, skipping the first 10
    >>> q = From(Users) >> Order(Get.date_joined) >> Limit(100, offset=10) >> Select(Get.id, Get.name)
    ```
    """

    limit: Optional[int]
    offset: Optional[int]
    over: Optional[SQLNode]

    def __init__(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        over: Optional[SQLNode] = None,
    ) -> None:
        super().__init__()
        self.limit = limit
        self.offset = offset
        self.over = over

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(
            self.limit, offset=self.offset, over=_rebase_node(self.over, pre)
        )

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "Limit"
        args = []

        if self.limit is not None:
            args.append(to_doc(self.limit))
        if self.offset is not None:
            args.append(assg_expr("offset", to_doc(self.offset)))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


class Lit(SQLNode):
    """
    `Lit` node represents a SQL literal value. FunSQL automatically casts values of the
    regular python datatypes to a `Lit` node, when a node value is expected.

    The set of literal types supported by FunSQl include:
    * int
    * float
    * str
    * bool
    * None - translated to a NULL literal
    * datetime.date
    * datetime.time
    * datetime.datetime - translated to a TIMESTAMP
    * datetime.timedelta - translated to an INTERVAL value

    NOTE: Some SQL engines might not support all the literal types listed here, for example,
    SQLite works with date/time values using functions rather than specific data types. In
    these cases, FunSQL will raise an error when serializing the query.

    Args:
        val: literal value

    Examples:
    ```python
    >>> q = Select(1, Lit(2), "hello", Lit("world"), None, Lit(None))
    ```
    """

    val: Any

    def __init__(self, val: Any) -> None:
        super().__init__()
        assert isinstance(
            val, LITERAL_TYPES
        ), f"Unexpected object of type: {type(val)} being cast as a literal"
        self.val = val

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        val = to_doc(self.val)
        return f"Lit({val})"


class Order(TabularNode):
    """
    `Order` sorts the results of the input subquery by the `sort key`, specified
    as a set of columns. It is translated to an `ORDER BY` clause in the final SQL
    query.

    Each column reference can be passed into a `Sort` node to specify the sort order
    for it - ascending/descending.

    Args:
        *by: column reference to sort by

    Examples:
    ```python
    # fetch students with their birthdays early in the year
    >>> q = (
            From(Students)
            >> Order(Fun.extract("MONTH", Get.date_of_birth))
            >> Limit(10)
            >> Select(Get.name, Get.date_of_birth)
        )

    # sort error logs by their severity level and reported time, critical ones first
    >>> q = (
            From(ErrorLogs)
            >> Order(Get.severity_level >> Desc(), Get.timestamp)
            >> Limit(10)
        )
    ```
    """

    by: list[SQLNode]
    over: Optional[SQLNode]

    def __init__(
        self,
        *by: SQLNode,
        over: Optional[SQLNode] = None,
    ) -> None:
        super().__init__()
        self.by = [_cast_to_node(arg) for arg in by]
        self.over = _cast_to_node_skip_none(over)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(*self.by, over=_rebase_node(self.over, pre))

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "Order"
        args = []

        if len(self.by) == 0:
            args.append(assg_expr("by", to_doc("[]")))
        else:
            for arg in self.by:
                args.append(to_doc(arg, ctx))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


class Partition(TabularNode):
    """
    A `Partition` node bunches together the rows of the input subquery by the `partition key`,
    which is a set of columns; not unlike a Group node. However, instead of splitting the dataset,
    each partition provides a `frame` for the rows in it. Aggregate columns can then be computed
    for the partition, the aggregation for each row is only done over the rows related to it, via
    sharing the same partition `frame`.

    Args:
        *by: column reference to partition by
        order_by: columns to order the rows in each frame by
        frame: argument to further customize the extent of rows related to a row

    Examples:
    ```python
    # calculate country totals with state subtotals already calculated
    >>> q = (
            From(Sales)
            >> Group(Get.country, Get.state)
            >> Select(Get.country, Get.state, Agg.sum(Get.amount) >> As("state_total"))
            >> Partition(Get.country)
            >> Select(Get.country, Get.state, Agg.sum(Get.state_total) >> As("country_total"))
        )

    # calculate the moving average of customer tickets each day
    >>> q = (
            From(Feedback)
            >> Group(Get.date)
            >> Select(Get.date, Agg.avg(Get.tickets) >> As("tickets_avg"))
            >> Partition(order_by=[Get.date],
                         frame=Frame(F.ROWS, F.pre(3), F.follow(3)))
            >> Select(Get.date, Agg.avg(Get.tickets_avg) >> As("tickets_moving_avg"))
        )
    ```
    """

    by: list[SQLNode]
    order_by: list[SQLNode]
    frame: Optional[Frame]
    over: Optional[SQLNode]

    def __init__(
        self,
        *by: SQLNode,
        order_by: Optional[list[SQLNode]] = None,
        frame: Optional[Frame] = None,
        over: Optional[SQLNode] = None,
    ) -> None:
        super().__init__()
        self.by = [_cast_to_node(arg) for arg in by]
        self.order_by = (
            [] if order_by is None else [_cast_to_node(arg) for arg in order_by]
        )
        self.frame = frame
        self.over = _cast_to_node_skip_none(over)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(
            *self.by,
            order_by=self.order_by,
            frame=self.frame,
            over=_rebase_node(self.over, pre),
        )

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "Partition"
        args = []

        for a in self.by:
            args.append(to_doc(a, ctx))
        if len(self.order_by) > 0:
            order_args = list_expr([to_doc(a, ctx) for a in self.order_by])
            args.append(assg_expr("order_by", order_args))
        if self.frame is not None:
            args.append(assg_expr("frame", to_doc(self.frame)))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


class Select(TabularNode):
    """
    `Select` defines an ordered set of columns to keep from the input subquery. Each
    column reference must have a unique alias, set explicity using the `As` node or inferred
    by FunSQL.

    Args:
        *args: column references to select

    Examples:
    ```python
    # get the number of flights out of each airport
    >>> q = From(Flights) >> Group(Get.origin) >> Select(Get.origin, Agg.count() >> As("count"))

    # get list of students and their ages
    >>> q = From(Students) >> Select(Get.name, Fun("-", Get.date_of_birth, Fun.now()) >> As("age"))
    ```
    """

    args: list[SQLNode]
    over: Optional[SQLNode]
    label_map: dict[Symbol, int]

    def __init__(
        self,
        *args: NODE_MATERIAL,
        over: Optional[SQLNode] = None,
    ) -> None:
        super().__init__()
        self.args = [_cast_to_node(arg) for arg in args]
        self.over = _cast_to_node_skip_none(over)
        self.label_map = populate_label_map(self, self.args)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(*self.args, over=_rebase_node(self.over, pre))

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "Select"
        args = []

        if len(self.args) == 0:
            args.append(assg_expr("args", to_doc("[]")))
        else:
            for arg in self.args:
                args.append(to_doc(arg, ctx))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


class Sort(SQLNode):
    """
    `Sort` node indicates the sorting order for the values in a column, when sorting a
    dataset on it. It is used with the `Order` or `Partition` nodes.

    `Asc` and `Desc` are functions to create instances of `Sort` node directly.

    Args:
        value: sort order ascending/descending?
        nulls: are the null values placed first/last?

    Examples:
    ```python
    # sort patients in decreasing order by their age, alphabetically
    >>> q = From(Patients) >> Order(Get.date_of_birth >> Asc(), Get.name) >> Limit(100)
    ```
    """

    value: ValueOrder
    nulls: Optional[NullsOrder]
    over: Optional[SQLNode]

    def __init__(
        self,
        value: ValueOrder,
        nulls: Optional[NullsOrder] = None,
        over: Optional[SQLNode] = None,
    ) -> None:
        super().__init__()
        self.value = value
        self.nulls = nulls
        self.over = _cast_to_node_skip_none(over)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(
            value=self.value, nulls=self.nulls, over=_rebase_node(self.over, pre)
        )

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "Sort"
        args = []

        args.append(self.value.value)
        if self.nulls is not None:
            args.append(assg_expr("nulls", self.nulls.value))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


class Var(SQLNode, metaclass=ExtendAttrsFull):
    """
    `Var` node creates a placeholder for a query parameter. The value for the placeholder
    variable can also be provided using a `Bind` node, either as a SQL literal, or as a
    column reference, in which case it gets rendered as a correlated query/lateral join.
    Refer to the `Bind` node for details.

    Examples:
    ```python
    # get the capital for a country
    >>> q = From(Countries) >> Where(eq(Get.name, Var.COUNTRY_NAME)) >> Select(Get.name, Get.capital_name)
    ```
    """

    _name: Symbol

    def __init__(self, name: Union[str, Symbol]) -> None:
        super().__init__()
        self._name = S(name)

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        return f"Var.{self._name}"


class Where(TabularNode):
    """
    Filter the input dataset with the given condition, returning a subset of the rows in it.

    Args:
        condition: an input/derived column of boolean type, rows where it is false are filtered out

    Examples:
    ```python
    # fetch names of students playing basketball
    >>> q = (
            From(Students)
            >> Join(From(Activities) >> As("activity"), on=eq(Get.student_id, Get.id))
            >> Where(Fun("=", Get.activity.name, "basketball"))
            >> Select(Get.name)
        )
    ```
    """

    condition: SQLNode
    over: Optional[SQLNode]

    def __init__(
        self, condition: NODE_MATERIAL, over: Optional[SQLNode] = None
    ) -> None:
        super().__init__()
        self.condition = _cast_to_node(condition)
        self.over = _cast_to_node_skip_none(over)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(
            condition=self.condition, over=_rebase_node(self.over, pre)
        )

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "Where"
        args = [to_doc(self.condition, ctx)]
        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


class With(TabularNode):
    """
    `With` node is used to create a CTE(common table expression). For larger queries,
    abstracting out a subquery that is used multiple times is helpful.

    The subquery wrapped using a `With` node can be accessed in its *parent* node, specified
    using the `over` argument. Each subquery must be aliased using an `As` node, so it can be
    referenced by the alias.

    Args:
        args: subqueries to be wrapped in a CTE, multiple can be specified.

    Examples:
    ```python
    # fetch the names of students playing basketball
    >>> q = (
            From(Students)
            >> Where(Fun.in(Get.id, From(S.bb_players) >> Select(Get.student_id)))
            >> With(From("Activities") >> Where(Fun("=", Get.name, "basketball")) >> As("bb_players"))
        )
    """

    args: list[SQLNode]
    materialized: Optional[bool]
    label_map: dict[Symbol, int]
    over: Optional[SQLNode]

    def __init__(
        self,
        *args: SQLNode,
        materialized: Optional[bool] = None,
        over: Optional[SQLNode] = None,
    ) -> None:
        super().__init__()
        self.args = [_cast_to_node(arg) for arg in args]
        self.materialized = materialized
        self.over = _cast_to_node_skip_none(over)
        self.label_map = populate_label_map(self, self.args)

    def rebase(self, pre: SQLNode) -> "SQLNode":
        return self.__class__(
            *self.args,
            materialized=self.materialized,
            over=_rebase_node(self.over, pre),
        )

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "With"
        args = []

        if len(self.args) == 0:
            args.append(assg_expr("args", to_doc("[]")))
        else:
            for a in self.args:
                args.append(to_doc(a, ctx))
        if self.materialized is not None:
            args.append(assg_expr("materialized", to_doc(self.materialized)))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


# -----------------------------------------------------------
# utility functions to export
# -----------------------------------------------------------


def Asc(nulls: Optional[NullsOrder] = None) -> Sort:
    """Shorthand to create a Sort node with ASC direction"""
    return Sort(ValueOrder.ASC, nulls=nulls)


def Desc(nulls: Optional[NullsOrder] = None) -> Sort:
    """Shorthand to create a Sort node with DESC direction"""
    return Sort(ValueOrder.DESC, nulls=nulls)


def aka(*args, **kwargs) -> SQLNode:
    """Shorthand to create an alias for a node. To be used as:

    >>> aka(Get.person, S.person_id)
    >>> aka(100, "count")
    >>> aka(count=100)
    >>> aka(sum=Agg.sum(Get.population))
    etc.
    """
    assert (
        len(args) == 0 or len(kwargs) == 0
    ), "only one of args or kwargs should be passed"

    if len(kwargs) == 1:
        name, node_like = next(iter(kwargs.items()))
    elif len(args) == 2:
        node_like, name = args
    else:
        raise Exception("Invalid arguments passed to the aliasing routine `aka`")
    return As(name=S(name), over=_cast_to_node(node_like))


def eq(left: SQLNode, right: SQLNode) -> SQLNode:
    """Shorthand to create a Function node checking equality"""
    return Fun("=", left, right)


def LeftJoin(
    joinee: SQLNode, on: NODE_MATERIAL, over: Optional[SQLNode] = None
) -> "Join":
    """Shorthand to create a left join node"""
    return Join(joinee=joinee, on=on, left=True, over=over)


class F:
    """Shorthand to pass arguments to window frame objects"""

    ROWS = FrameMode.ROWS
    RANGE = FrameMode.RANGE
    GROUPS = FrameMode.GROUPS
    EXCL_CURR = FrameExclude.CURRENT_ROW
    EXCL_GROUP = FrameExclude.GROUP
    EXCL_TIES = FrameExclude.TIES

    @staticmethod
    def curr_row() -> FrameEdge:
        return FrameEdge(FrameEdgeSide.CURRENT_ROW)

    @staticmethod
    def pre(val: Union[None, Any] = None) -> FrameEdge:
        return FrameEdge(FrameEdgeSide.PRECEDING, val)

    @staticmethod
    def follow(val: Union[None, Any] = None) -> FrameEdge:
        return FrameEdge(FrameEdgeSide.FOLLOWING, val)


# -----------------------------------------------------------
# label method implemented for nodes
# -----------------------------------------------------------


@register_union_type(label)
def _(node: Union[Agg, Fun, Get, Var]) -> Symbol:
    return node._name


@label.register
def _(node: As) -> Symbol:
    return node.name


@label.register
def _(node: Append) -> Symbol:
    lbl = label(node.over)
    if all(label(arg) == lbl for arg in node.args):
        return lbl
    else:
        return S("union")


@label.register
def _(node: From) -> Symbol:
    if isinstance(node.source, SQLTable):
        return node.source.name
    elif isinstance(node.source, Symbol):
        return node.source
    elif isinstance(node.source, dict):
        return S("values")
    else:
        return label(None)


@label.register
def _(node: Lit) -> Symbol:
    return label(None)


@register_union_type(label)
def _(node: Union[Bind, Define, Group, Iterate, Join, Limit, Order]) -> Symbol:
    return label(node.over)


@register_union_type(label)
def _(node: Union[Partition, Select, Sort, Where, With]) -> Symbol:
    return label(node.over)



"""
Constructs that provide the context to evaluate a query.
"""

from dataclasses import dataclass
from enum import Enum
from typing import Any, Iterator, Optional

from .common import S, Symbol
from .prettier import (
    Doc,
    QuoteContext,
    call_expr,
    assg_expr,
    list_expr,
    resolve,
    to_doc,
)

__all__ = [
    "SQLTable",
    "SQLCatalog",
    "ValuesTable",
    "VarStyle",
    "LimitStyle",
    "SQLDialect",
]


class SQLTable:
    name: Symbol
    columns: list[Symbol]
    schema: Optional[Symbol]

    def __init__(
        self, name: Symbol, columns: list[Symbol], schema: Optional[Symbol] = None
    ) -> None:
        self.name = S(name)
        self.columns = [S(col) for col in columns]
        self.schema = None if schema is None else S(schema)

    def __repr__(self) -> str:
        ctx = QuoteContext()
        return resolve(to_doc(self, ctx), 80)

    def pretty_repr(self, ctx: QuoteContext) -> Doc:
        name = "SQLTable"
        args = []

        args.append(str(self.name))
        if self.schema is not None:
            args.append(assg_expr("schema", str(self.schema)))
        if ctx.limit:
            args.append("...")
        else:
            args.append(
                assg_expr("columns", list_expr([str(col) for col in self.columns]))
            )
        return call_expr(name, args)


class ValuesTable:
    """To represent queries of the type - `FROM (VALUES ...) AS t(col1, col2, ...)`"""

    columns: tuple[Symbol]
    data: list[tuple]

    def __init__(self, columns: tuple[str], data: list[tuple]) -> None:
        for row in data:
            assert len(row) == len(columns)
        self.columns = tuple(S(c) for c in columns)
        self.data = data


class SQLCatalog:
    """SQL catalog to capture the structure of a table like object for constructing queries"""

    tables: dict[Symbol, SQLTable]
    dialect: "SQLDialect"

    def __init__(
        self, dialect: "SQLDialect", tables: Optional[dict[Symbol, SQLTable]] = None
    ) -> None:
        self.dialect = dialect
        self.tables = {} if tables is None else tables

    def __repr__(self) -> str:
        ctx = QuoteContext()
        return resolve(to_doc(self, ctx), 80)

    def pretty_repr(self, ctx: QuoteContext) -> Doc:
        name = "SQLCatalog"
        args = []

        args.append(assg_expr("dialect", str(self.dialect)))
        for t_name, table in self.tables.items():
            args.append(assg_expr(str(t_name), to_doc(table, ctx)))
        return call_expr(name, args)

    def get(self, key: Symbol) -> Optional[SQLTable]:
        """get the table with the given name"""
        return self.tables.get(key, None)

    def __getitem__(self, key: Symbol) -> SQLTable:
        """get the table with the given name"""
        return self.tables[key]

    def __len__(self) -> int:
        return len(self.tables)

    def __iter__(self) -> Iterator[tuple[Symbol, SQLTable]]:
        for name, table in self.tables.items():
            yield name, table


class VarStyle(Enum):
    NAMED = 1
    NUMBERED = 2
    POSITIONAL = 3


class LimitStyle(Enum):
    REGULAR = 1
    FETCH_FIRST_KIND = 2


@dataclass(repr=False)
class SQLDialect:
    name: str = "default"
    var_style: VarStyle = VarStyle.NAMED
    var_prefix: str = "?"
    id_quotes: tuple[str, str] = ('"', '"')
    has_bool_literals: bool = True
    limit_style: LimitStyle = LimitStyle.REGULAR
    has_recursive_annotation: bool = True
    has_as_columns: bool = True
    has_datetime_types: bool = True
    values_row_constructor: Optional[str] = None
    values_column_prefix: Optional[str] = "column"
    values_column_index: int = 1

    def __repr__(self) -> str:
        return f"SQLDialect(:{self.name})"



"""
This module implements a `Printer` object to print the FunSQL query 
structure, since they can get long and unwieldy. Adapted from:
* https://github.com/dtolnay/prettyplease/
* https://github.com/stevej2608/oppen-pretty-printer

both of which implement the algorithm first implemented by Derek C. Oppen. 
"""


import io
from enum import Enum
from collections import deque
from typing import Union, NamedTuple


SIZE_INFINITY = 1000_000

# -----------------------------------------------------------
# primitives to assmeble an expression to print
# -----------------------------------------------------------

Token = Union[str, "Break", "Begin", "End"]


class GroupBreak(Enum):
    fits = 1
    consistent = 2
    inconsistent = 3


class Break:
    blanks: int
    offset: int

    def __init__(self, blanks: int = 0, offset: int = 0) -> None:
        self.blanks = blanks
        self.offset = offset


class Begin:
    """
    Setting `ns_indent` to True overrides the offset for that `Begin...End`
    block to the cursor position at start. So, we can get formatting like:
    ```
    begin
         x = 1
    end
    ```
    instead of the regular (assuming, offset = 2):
    ```
    begin
      x = 1
    end
    ```
    with standard 2 or 4 space indentation.
    """

    offset: int
    break_typ: GroupBreak
    ns_indent: bool  # non-standard indent

    def __init__(
        self,
        offset: int = 0,
        break_typ: GroupBreak = GroupBreak.consistent,
        ns_indent: bool = False,
    ) -> None:
        self.offset = offset
        self.break_typ = break_typ
        self.ns_indent = ns_indent


class End:
    pass


# -----------------------------------------------------------
# The `Printer` object which does the work
# -----------------------------------------------------------


class PrintFrame(NamedTuple):
    break_typ: GroupBreak
    offset: int = 0


class TokenEntry:
    token: Token
    size: int

    def __init__(self, tok: Token, size: int) -> None:
        self.token = tok
        self.size = size


class Printer:
    _buf_offset: int  # incremented on each `advance_left` operation
    buffer: deque[TokenEntry]  # ringbuffer of tokens along with their calculated sizes
    left_total: int  # sum of size of tokens already printed
    right_total: int  # sum of size of tokens printed or queued
    scan_stack: deque[int]  # tracks the index of delimiter tokens in the buffer
    writer: "TokenWriter"

    def __init__(self, line_width: int = 80) -> None:
        self._buf_offset = 0
        self.buffer = deque()
        self.left_total = 0
        self.right_total = 0
        self.scan_stack = deque()
        self.writer = TokenWriter(line_width)

    @property
    def buffer_idx_last(self) -> int:
        return self._buf_offset + len(self.buffer) - 1

    @property
    def buffer_idx_first(self) -> int:
        return self._buf_offset

    def scan(self, token: Token):
        if isinstance(token, str):
            self.scan_str(token)
        elif isinstance(token, Begin):
            self.scan_begin(token)
        elif isinstance(token, End):
            self.scan_end(token)
        elif isinstance(token, Break):
            self.scan_break(token)
        else:
            raise ValueError(f"unexpected token type: {type(token)}")

    def scan_begin(self, token: Begin):
        if len(self.scan_stack) == 0:
            self.left_total = 1
            self.right_total = 1
            self.buffer.clear()

        entry = TokenEntry(token, -self.right_total)
        self.buffer.append(entry)
        self.scan_stack.append(self.buffer_idx_last)

    def scan_end(self, token: End):
        if len(self.scan_stack) == 0:
            self.writer.print(token, 0)
        else:
            entry = TokenEntry(token, -1)
            self.buffer.append(entry)
            self.scan_stack.append(self.buffer_idx_last)

    def scan_break(self, token: Break):
        if len(self.scan_stack) == 0:
            self.left_total = 1
            self.right_total = 1
            self.buffer.clear()

        self.check_stack(0)
        entry = TokenEntry(token, -self.right_total)
        self.buffer.append(entry)
        self.scan_stack.append(self.buffer_idx_last)
        self.right_total += token.blanks

    def scan_str(self, token: str):
        size = len(token)
        if len(self.scan_stack) == 0:
            self.writer.print(token, size)
        else:
            entry = TokenEntry(token, size)
            self.buffer.append(entry)
            self.right_total += size
            self.check_stream()

    def eof(self) -> None:
        """Resolve any outstanding tokens in the buffer, and push them to the writer"""
        if len(self.scan_stack) > 0:
            self.check_stack(0)
            self.advance_left()

    def getvalue(self) -> str:
        return self.writer.getvalue()

    def advance_left(self):
        while self.buffer[0].size >= 0:
            entry = self.buffer.popleft()
            self._buf_offset += 1

            self.writer.print(entry.token, entry.size)
            if isinstance(entry.token, str):
                assert len(entry.token) == entry.size
                self.left_total += entry.size
            elif isinstance(entry.token, Break):
                self.left_total += entry.token.blanks

            if len(self.buffer) == 0:
                break

    def check_stack(self, depth: int):
        """Takes elements off the scan stack if their sizes are known.
        This method gets called on getting a `Break` token, or at the end.
        """
        assert depth >= 0
        while len(self.scan_stack) > 0:
            idx = self.scan_stack[-1]
            entry = self.buffer[idx - self.buffer_idx_first]

            if isinstance(entry.token, Begin):
                # Begin -> ..., nothing to do here
                if depth == 0:
                    break
                else:
                    # Begin -> (End) -> ..., so we can resolve the Begin token
                    self.scan_stack.pop()
                    entry.size += self.right_total
                    depth -= 1
            elif isinstance(entry.token, End):
                # End -> ..., recurse to resolve the Begin token
                self.scan_stack.pop()
                entry.size = 1
                depth += 1
            elif isinstance(entry.token, Break):
                # Break -> ..., resolve the previous Break token
                self.scan_stack.pop()
                entry.size += self.right_total
                if depth == 0:
                    break
            else:
                raise Exception("unreachable case")

    def check_stream(self):
        """Check if the current line has gone beyond the available width"""
        while self.right_total - self.left_total > self.writer.space:
            if len(self.scan_stack) > 0:
                if self.buffer_idx_first == self.scan_stack[0]:
                    self.scan_stack.popleft()
                    self.buffer[0].size = SIZE_INFINITY

            self.advance_left()
            if len(self.buffer) == 0:
                break


class TokenWriter:
    output: io.StringIO
    print_stack: list[PrintFrame]
    line_width: int
    space: int
    indent: int

    def __init__(self, line_width) -> None:
        self.output = io.StringIO()
        self.print_stack = []
        self.line_width = line_width
        self.space = line_width
        self.indent = 0

    def print(self, tok: Token, size: int):
        """Called by the Scanner object to an output buffer"""
        if isinstance(tok, str):
            self.output.write(tok)
            self.space -= size

        elif isinstance(tok, Begin):
            if size > self.space:
                # record the current indentation for recovery later
                self.print_stack.append(PrintFrame(tok.break_typ, self.indent))
                if tok.ns_indent:
                    self.indent = self.line_width - self.space
                else:
                    self.indent = self.indent + tok.offset
            else:
                self.print_stack.append(PrintFrame(GroupBreak.fits, 0))

        elif isinstance(tok, End):
            break_typ, indent = self.print_stack.pop()
            if break_typ != GroupBreak.fits:
                self.indent = indent  # restore the original indent level

        elif isinstance(tok, Break):
            top = self.print_stack[-1]
            fits_line = top.break_typ == GroupBreak.fits or (
                top.break_typ == GroupBreak.inconsistent and size <= self.space
            )
            if fits_line:
                self.space -= tok.blanks
                self.output.write(" " * tok.blanks)
            else:
                self.output.write("\n")
                offset = self.indent + tok.offset
                self.output.write(" " * offset)
                self.space = self.line_width - offset

        else:
            raise ValueError(f"Unknown token type: {type(tok)}")

    def getvalue(self) -> str:
        self.output.flush()
        return self.output.getvalue()



"""
This module implements schema introspection for the common SQL databases, so we 
don't have to construct the SQLCatalog by hand. 
"""

from typing import Optional

from ..common import S
from ..sqlcontext import SQLTable

__all__ = [
    "reflect_default",
    "reflect_sqlite",
    "reflect_mysql",
    "reflect_postgres",
    "make_sql_tables",
]


REFLECT_POSTGRES = """
SELECT
    n.nspname AS schema_,
    c.relname AS name_,
    a.attname AS column_
FROM
    pg_catalog.pg_namespace AS n
JOIN 
    pg_catalog.pg_class AS c 
    ON c.relnamespace = n.oid
JOIN
    pg_catalog.pg_attribute AS a
    ON a.attrelid = c.oid
WHERE
    n.nspname = {schema}
    AND 
    c.relkind IN ('r', 'v')
    AND
    a.attnum > 0
    AND
    NOT a.attisdropped
ORDER BY
    n.nspname, 
    c.relname, 
    a.attnum;
"""


REFLECT_SQLITE = """
SELECT 
    null AS schema_, 
    sm.name AS name_, 
    pti.name AS column_
FROM 
    sqlite_master AS sm
JOIN 
    pragma_table_info(sm.name) AS pti
    ON TRUE
WHERE 
    sm.type IN ('table', 'view') 
    AND 
    NOT sm.name LIKE 'sqlite_%'
ORDER BY 
    sm.name, 
    pti.cid;
"""


REFLECT_DEFAULT = """
SELECT
    c.table_schema AS schema_, 
    c.table_name AS name_,
    c.column_name AS column_
FROM 
    information_schema.columns AS c
WHERE
    c.table_schema = {schema}
ORDER BY
    c.table_schema,
    c.table_name,
    c.ordinal_position;
"""


def reflect_default(schema: str) -> str:
    return REFLECT_DEFAULT.format(schema=f"'{schema}'")


def reflect_postgres(schema: str = "public") -> str:
    return REFLECT_POSTGRES.format(schema=f"'{schema}'")


def reflect_sqlite() -> str:
    return REFLECT_SQLITE


def reflect_mysql(schema: Optional[str] = None) -> str:
    return REFLECT_DEFAULT.format(
        schema="DATABASE()" if schema is None else f"'{schema}'"
    )


def make_sql_tables(
    reflect_result: list[tuple[Optional[str], str, str]]
) -> list[SQLTable]:
    """Given a list of tuples of (schema, table, column), return a list of
    SQLTable objects.
    """
    tables_to_cols: dict[tuple[Optional[str], str], list[str]] = dict()
    for schema, table, column in reflect_result:
        if (schema, table) not in tables_to_cols:
            tables_to_cols[(schema, table)] = []
        tables_to_cols[(schema, table)].append(column)

    sql_tables: list[SQLTable] = []
    for (schema, table), columns in tables_to_cols.items():
        sql_tables.append(
            SQLTable(
                name=S(table),
                columns=[S(c) for c in columns],
                schema=S(schema) if schema is not None else None,
            )
        )
    return sql_tables



"""
This module implements primitives to render FunSQL objects into a stream 
of tokens for pretty printing. The pretty printing algorithm used is the 
one first implemented by Derek C. Oppen. 
"""


import datetime
import os
from functools import singledispatch
from contextlib import contextmanager
from typing import Union, Optional, Any, Callable, Generator

from .common import Symbol, register_union_type
from .prettyprint import *


Doc = Union[Token, list["Doc"]]

# -----------------------------------------------------------
# utility routines to create/merge `Doc` objects
# -----------------------------------------------------------


def join_docs(sep: Callable[[], Doc], docs: list[Doc]) -> list[Doc]:
    """Join a sequence of tokens with a separator"""
    list_vals = []
    for i, doc in enumerate(docs):
        if i != 0:
            list_vals.append(sep())
        list_vals.append(doc)
    return list_vals


def assg_expr(key: str, val: "Doc") -> Doc:
    return [key, " = ", val]


def annotate_expr(key: str, val: "Doc") -> Doc:
    return [key, ": ", val]


def pipe_expr(left: "Doc", right: "Doc") -> Doc:
    return [Begin(ns_indent=True), left, " >>", Break(1), right, End()]


def highlight_expr(doc: "Doc") -> "Doc":
    highlight = ["^^", Break(blanks=2), doc, Break(blanks=2), "^^"]
    return [Begin(ns_indent=True), *highlight, End()]


def block_expr(args: list[Doc]) -> list[Doc]:
    assert isinstance(args, list), f"Expected list, got {type(args)}"
    sep = lambda: [",", Break(blanks=1)]
    return [Begin(ns_indent=True), join_docs(sep, args), End()]


def call_expr(name: str, args: list[Doc]) -> Doc:
    return [name, "(", block_expr(args), ")"]


def list_expr(args: list["Doc"]) -> Doc:
    return ["[", block_expr(args), "]"]


def full_query_expr(args: list["Doc"]) -> Doc:
    assert isinstance(args, list), f"Expected list, got {type(args)}"
    sep = lambda: [",", Break(blanks=10000)]
    return [
        Begin(),
        "let ",
        Begin(ns_indent=True),
        join_docs(sep, args),
        End(),
        Break(blanks=10000),
        "end",
        End(),
    ]


# -----------------------------------------------------------
# convert standard python datatypes to `Doc` objects
# -----------------------------------------------------------


@singledispatch
def to_doc(val: Any, ctx: Optional["QuoteContext"] = None) -> Doc:
    if not hasattr(val, "pretty_repr"):
        raise TypeError(f"Unsupported literal type for SQL expression: {val}")
    return val.pretty_repr(ctx)


@register_union_type(to_doc)
def _(val: Union[int, float, bool]) -> str:
    return str(val)


@to_doc.register
def _(val: str) -> str:
    return f'"{val}"'


@to_doc.register
def _(val: Symbol) -> str:
    return str(val)


@to_doc.register
def _(val: datetime.date) -> str:
    return f"DATE '{val.strftime('%Y-%m-%d')}'"


@to_doc.register
def _(val: datetime.time) -> str:
    return f"TIME '{val.strftime('%H:%M:%S')}'"


@to_doc.register
def _(val: datetime.datetime) -> str:
    return f"TIMESTAMP '{val.isoformat()}'"


@to_doc.register
def _(val: datetime.timedelta) -> str:
    return f"INTERVAL '{val.total_seconds()} SECONDS'"


@to_doc.register
def _(val: None) -> str:
    return "NULL"


# -----------------------------------------------------------
# print out a `Doc` object
# -----------------------------------------------------------


def flatten(doc: list["Doc"]) -> Generator[Token, None, None]:
    """Flatten a Doc object into a sequence of tokens"""
    for item in doc:
        if isinstance(item, list):
            yield from flatten(item)
        else:
            yield item


def get_screen_width() -> int:
    """Get the current screen width"""
    try:
        sz = os.get_terminal_size().columns
        return int(sz * 5 / 6)
    except:
        return 80


def resolve(doc: "Doc", width: int) -> str:
    """Resolve a Doc object into a string"""
    printer = Printer(width)
    if not isinstance(doc, list):
        doc = [doc]
    for tok in flatten(doc):
        printer.scan(tok)
    printer.eof()
    return printer.getvalue()


class QuoteContext:
    """Auxiliary information for pretty printing"""

    limit: bool  # indicate if the child objects don't need to be expanded
    vars_: dict[Any, Symbol]  # substitute node/clause objects with equivalent symbols

    def __init__(self, limit: bool = False, vars_: Optional[dict[Any, Symbol]] = None):
        self.limit = limit
        self.vars_ = vars_ if vars_ is not None else {}

    @contextmanager
    def set(self, **kwargs):
        """Temporarily set the QuoteContext attributes"""
        prev = {k: getattr(self, k) for k in kwargs}
        for k, v in kwargs.items():
            setattr(self, k, v)

        try:
            yield
        finally:
            for k, v in prev.items():
                setattr(self, k, v)



# constructing a query
from .nodedefs import *  # type: ignore
from .nodes import SQLNode, TabularNode, ErrDuplicateLabel, ErrIllFormed
from .clausedefs import (
    Frame,
    FrameMode,
    FrameEdgeSide,
    FrameEdge,
    FrameExclude,
    ValueOrder,
    NullsOrder,
)
from .common import Symbol, S

# compiling queries
from .render import (
    SQLString,
    RenderDepth,
    render,
    render_clause,
    dialect_default,
)
from .sqlcontext import SQLCatalog, SQLDialect, SQLTable



from functools import lru_cache
from typing import Any, Generic, Union, Optional, Type, TypeVar, get_type_hints

from .common import Symbol
from .prettier import Doc, QuoteContext, resolve, to_doc

__all__ = ["SQLClause"]


def _get_origin(typ: Any) -> Any:
    # HACK: looks very fragile, but `get_origin`, `get_args` are not present in python <3.8
    return typ.__origin__ if hasattr(typ, "__origin__") else None


def _get_args(typ: Any) -> tuple:
    return typ.__args__ if hasattr(typ, "__args__") else tuple()


@lru_cache
def _get_refs_for_clause(
    object_typ: Type["SQLClause"],
) -> tuple[list[str], list[str]]:
    """Compute list of attributes on a SQLClause class that are scalars vs sequences."""

    assert issubclass(object_typ, SQLClause)

    scalar_attrs, sequence_attrs = [], []
    for attr, typ in get_type_hints(object_typ).items():
        container = _get_origin(typ)
        if container == list:
            sequence_attrs.append(attr)
        elif container == Union:
            args = _get_args(typ)
            assert (
                len(args) == 2 and type(None) in args
            ), "TODO: accomodate union types other than optional"
            arg = [x for x in args if x != type(None)][0]
            if _get_origin(arg) == list:
                sequence_attrs.append(attr)
            else:
                scalar_attrs.append(attr)
        else:
            scalar_attrs.append(attr)
    return scalar_attrs, sequence_attrs


T = TypeVar("T", bound="SQLClause")


class SQLClause(Generic[T]):
    """Base class for a SQL clause, that translates directly to an SQL expression"""

    def __repr__(self) -> str:
        return resolve(to_doc(self), 80)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> Doc:
        raise NotImplementedError(
            f"pretty_repr method isn't implemented on the SQLClause class - {type(self)}"
        )

    def rebase(self: T, *args, **kwargs) -> T:
        raise NotImplementedError(
            f"rebase method isn't implemented on the SQLClause class - {type(self)}"
        )

    def __rshift__(self, other):
        """
        We should implement __rrshift__ instead but reflected operators in python
        don't work for objects of the same type, so this indirection.
        [link](https://docs.python.org/3/reference/datamodel.html?highlight=radd#id11)
        """
        if isinstance(other, SQLClause):
            return other.rebase(self)
        else:
            raise NotImplementedError(
                f">> isn't a valid operation with a SQLClause and {type(other)}"
            )

    def __rrshift__(self, other: Symbol) -> "SQLClause":
        if isinstance(other, Symbol):
            return self.rebase(other)
        else:
            raise NotImplementedError(
                f">> isn't a valid operation with a SQLClause and {type(other)}"
            )

    def _key(self):
        """We want to implement hashing for SQLClause objects, since they are compared
        to deduplicate clauses in the rendered SQL.

        TODO: This is fairly expensive, and increases render time by like 50pc. Should we
        instead call a custom function when clause equality is really evaluated, instead
        of implementing `__eq__`?

        Credits: The pattern is lifted off a stackoverflow answer I can't find the link to.
        """
        scalar_attrs, sequence_attrs = _get_refs_for_clause(self.__class__)
        vals = []
        for attr in scalar_attrs:
            vals.append(getattr(self, attr))
        for attr in sequence_attrs:
            val = getattr(self, attr)
            vals.append(tuple(val) if val is not None else None)
        return tuple(vals)

    def __hash__(self) -> int:
        return hash(self._key())

    def __eq__(self, other: Any) -> bool:
        if type(self) is not type(other):
            return NotImplemented
        return self._key() == other._key()



"""
This module implements the compiler pass to translate the tree of SQLnode 
objects to a tree of SQLClause objects, which represents the lexical structure 
of the SQL query. The SQLClause expression can then be serialized almost 
directly into the SQL query string. 
"""

from contextlib import contextmanager
from functools import singledispatch
from typing import Optional, Union, overload

from ..clausedefs import *
from ..clauses import *
from ..nodedefs import *
from ..nodes import *
from ..sqlcontext import *
from .annotate import *
from .types import *

__all__ = ["translate_toplevel", "TranslateContext"]

NODE_TO_CLAUSE = dict[SQLNode, SQLClause]
NODE_TO_ALIAS = dict[SQLNode, Symbol]
ALIAS_TO_CLAUSE = dict[Symbol, SQLClause]


class Assemblage:
    """Represents a partially constructed SQL query"""

    clause: Optional[SQLClause]  # An SQL subquery, with(out) SELECT args
    cols: ALIAS_TO_CLAUSE  # SELECT args if required
    repl: NODE_TO_ALIAS  # mapping referenced nodes to column aliases

    def __init__(
        self,
        clause: Optional[SQLClause],
        cols: Optional[ALIAS_TO_CLAUSE] = None,
        repl: Optional[NODE_TO_ALIAS] = None,
    ) -> None:
        self.clause = clause
        self.cols = cols if cols is not None else dict()
        self.repl = repl if repl is not None else dict()

    def get_complete_select(self) -> SQLClause:
        """Construct the SELECT clause to a partially assembled query if it didn't have one
        and return.
        """
        clause = self.clause
        if not isinstance(clause, (SELECT, UNION)):
            args = cols_to_select_args(self.cols)
            clause = SELECT(*args, over=clause)
        return clause

    def get_make_subs(self, alias: Optional[Symbol]) -> NODE_TO_CLAUSE:
        """Build and return a node-to-clause mapping assuming the assemblage will be
        extended/completed, if the input `alias` arg is None/not None.
        """
        subs: NODE_TO_CLAUSE = dict()
        if alias is None:
            for ref, name in self.repl.items():
                subs[ref] = self.cols[name]
        else:
            cache: ALIAS_TO_CLAUSE = dict()
            for ref, name in self.repl.items():
                if name not in cache:
                    cache[name] = ID(name=name, over=ID(name=alias))
                subs[ref] = cache[name]
        return subs

    def unwrap_repl(self) -> "Assemblage":
        repl_p: NODE_TO_ALIAS = dict()
        for ref, name in self.repl.items():
            assert isinstance(ref, NameBound)
            repl_p[ref.over] = name
        return Assemblage(self.clause, cols=self.cols, repl=repl_p)


def cols_to_select_args(cols: ALIAS_TO_CLAUSE) -> list[SQLClause]:
    """Pack the arguments to a SELECT clause; this routine goes over all the
    columns and assigns them aliases.
    """
    args = []
    for name, clause in cols.items():
        # skip if the column is already a legit SELECT arg
        if not (isinstance(clause, ID) and clause.name == name):
            clause = AS(name=name, over=clause)
        args.append(clause)
    if len(args) == 0:
        args.append(LIT(None))
    return args


def translates_to_repl_n_cols(
    translates: list[tuple[SQLNode, SQLClause]]
) -> tuple[NODE_TO_ALIAS, ALIAS_TO_CLAUSE]:
    """Build a map of nodes to aliases and SELECT columns"""
    repl: NODE_TO_ALIAS = dict()
    cols: ALIAS_TO_CLAUSE = dict()

    renames: dict[tuple[Symbol, SQLClause], Symbol] = dict()
    counter: dict[Symbol, int] = dict()
    for ref, clause in translates:
        name = label(ref)
        counter[name] = counter.get(name, 0) + 1
        k = counter[name]

        # TODO: this is a trouble site, until we implement eq/hash for *all* SQLClause objects
        if k > 1 and (name, clause) in renames:
            # filters out duplicate columns
            name_p = renames[(name, clause)]
            repl[ref] = S(name_p)
        else:
            name_p = S(f"{name}_{k}") if k > 1 else name
            repl[ref] = name_p
            cols[name_p] = clause
            renames[(name, clause)] = name_p

    return repl, cols


def refs_to_repl_n_cols(refs: list[SQLNode]) -> tuple[NODE_TO_ALIAS, ALIAS_TO_CLAUSE]:
    """Build a map of nodes to aliases and implicit SELECT columns for a UNION query"""
    repl: NODE_TO_ALIAS = dict()
    cols: ALIAS_TO_CLAUSE = dict()

    # TODO: make sure if we really want a sequence like - q, q_2, q_3, ...? Or something else.
    counter: dict[Symbol, int] = dict()
    for ref in refs:
        name = label(ref)
        counter[name] = counter.get(name, 0) + 1
        k = counter[name]

        name_p = S(f"{name}_{k}") if k > 1 else name
        repl[ref] = name_p
        cols[name_p] = ID(name=name_p)

    return repl, cols


class CTEAssemblage:
    asmb: Assemblage
    name: Symbol
    schema: Optional[Symbol]
    materialized: Optional[bool]

    def __init__(
        self,
        asmb: Assemblage,
        name: Symbol,
        schema: Optional[Symbol] = None,
        materialized: Optional[bool] = None,
    ) -> None:
        self.asmb = asmb
        self.name = name
        self.schema = schema
        self.materialized = materialized


class TranslateContext:
    dialect: SQLDialect
    aliases: dict[Symbol, int]
    cte_map: dict[SQLNode, CTEAssemblage]
    recursive: bool
    vars_: ALIAS_TO_CLAUSE
    subs: NODE_TO_CLAUSE

    def __init__(self, ctx: AnnotateContext) -> None:
        self.dialect = ctx.catalog.dialect
        self.aliases = dict()
        self.cte_map = dict()
        self.recursive = False
        self.vars_ = dict()
        self.subs = dict()

    @contextmanager
    def substitute_vars_n_subs(
        self,
        *,
        vars_: Optional[ALIAS_TO_CLAUSE] = None,
        subs: Optional[NODE_TO_CLAUSE] = None,
    ):
        prev_vars_ = self.vars_
        prev_subs = self.subs

        try:
            if vars_ is not None:
                self.vars_ = vars_
            if subs is not None:
                self.subs = subs
            yield
        finally:
            self.vars_ = prev_vars_
            self.subs = prev_subs

    def allocate_alias(self, node_or_alias: Union[SQLNode, Symbol]) -> Symbol:
        """allocate a new alias"""
        if isinstance(node_or_alias, SQLNode):
            assert isinstance(node_or_alias, Box)
            alias = node_or_alias.typ.name
        else:
            alias = node_or_alias

        n = self.aliases.get(alias, 0) + 1
        self.aliases[alias] = n
        return S(f"{alias}_{n}")


def translate_toplevel(node: SQLNode, ctx: TranslateContext) -> SQLClause:
    """Translate the top-level SQLNode to produce a SQLClause representing the full query.

    * The `translate` routine starts from the query node, and produces a SQLClause, while
    any CTEs encountered are collected in the context.
    * The CTEs are converted to a WITH clause, and prepended to the previous output.
    """
    clause = translate(node, ctx)
    with_args: list[SQLClause] = []

    # The CTE nodes encountered are collected in the translation context object
    for cte_asmb in ctx.cte_map.values():
        cols = [S(name) for name in cte_asmb.asmb.cols]
        if len(cols) == 0:
            cols.append(S("_"))

        over = cte_asmb.asmb.get_complete_select()
        if cte_asmb.materialized is not None:
            over = NOTE(
                "MATERIALIZED" if cte_asmb.materialized else "NOT MATERIALIZED",
                over=over,
            )
        arg = AS(name=cte_asmb.name, columns=cols, over=over)
        with_args.append(arg)

    if len(with_args) > 0:
        clause = WITH(*with_args, over=clause, recursive=ctx.recursive)
    return clause


# -----------------------------------------------------------
# translating scalar nodes
# -----------------------------------------------------------


@overload
def translate(node: SQLNode, ctx: TranslateContext) -> SQLClause:
    ...


@overload
def translate(node: None, ctx: TranslateContext) -> None:
    ...


@overload
def translate(node: list[SQLNode], ctx: TranslateContext) -> list[SQLClause]:
    ...


def translate(
    node: Union[None, SQLNode, list[SQLNode]], ctx: TranslateContext
) -> Union[None, SQLClause, list[SQLClause]]:
    """translate a SQLNode"""
    if node is None:
        return None
    elif isinstance(node, list):
        return [translate(n, ctx) for n in node]

    assert isinstance(node, SQLNode), f"expected SQLNode, got: {type(node)}"
    if node in ctx.subs:
        return ctx.subs[node]
    elif isinstance(node, Box):
        base: Assemblage = assemble(node, ctx)
        return base.get_complete_select()
    else:
        return translate_scalar(node, ctx)


@singledispatch
def translate_scalar(node: SQLNode, ctx: TranslateContext) -> Optional[SQLClause]:
    """specific translations for each subtype of SQLNode"""
    raise NotImplementedError(
        f"Translation isn't implemented for node of type: {type(node)}"
    )


@translate_scalar.register
def _(node: Agg, ctx: TranslateContext) -> SQLClause:
    args: list[SQLClause] = []
    if str(node._name).upper() == "COUNT" and len(node.args) == 0:
        args = [OP(S("*"))]
    else:
        args = translate(node.args, ctx)
    filter_ = translate(node.filter_, ctx)
    return AGG(node._name, *args, distinct=node.distinct, filter_=filter_)


@translate_scalar.register
def _(node: As, ctx: TranslateContext) -> Optional[SQLClause]:
    return translate(node.over, ctx)


@translate_scalar.register
def _(node: IntBind, ctx: TranslateContext) -> Optional[SQLClause]:
    vars_p = ctx.vars_.copy()
    for name, i in node.label_map.items():
        vars_p[name] = translate(node.args[i], ctx)

    with ctx.substitute_vars_n_subs(vars_=vars_p):
        return translate(node.over, ctx)


_FUNC_REPLACE = {
    "==": "=",
    "!=": "<>",
}


@translate_scalar.register
def _(node: Fun, ctx: TranslateContext) -> SQLClause:
    fn = str(node._name).upper()

    if fn in ("NOT", "LIKE", "EXISTS", "=", "==", "!="):
        name = S(_FUNC_REPLACE.get(fn, fn))
        args = translate(node.args, ctx)
        return OP(name, *args)

    elif fn in ("AND", "OR"):
        default = True if fn == "AND" else False
        args = translate(node.args, ctx)
        if len(args) == 0:
            return LIT(default)
        elif len(args) == 1:
            return args[0]
        elif len(args) == 2 and isinstance(args[0], Lit) and args[0].val == default:
            return args[1]
        elif isinstance(args[0], OP) and str(args[0].name) == fn:
            # merge nested AND/OR
            return OP(args[0].name, *args[0].args, *args[1:])
        else:
            return OP(S(fn), *args)

    elif fn in ("IN", "NOT IN"):
        default = False if fn == "IN" else True
        if len(node.args) <= 1:
            return LIT(default)
        else:
            args = translate(node.args, ctx)
            if len(args) == 2 and isinstance(args[1], (SELECT, UNION)):
                return OP(S(fn), args[0], args[1])
            else:
                return OP(S(fn), args[0], FUN(S("_"), *args[1:]))

    elif fn in ("IS NULL", "IS NOT NULL"):
        name = S("IS" if fn == "IS NULL" else "IS NOT")
        return OP(name, *translate(node.args, ctx), None)

    elif fn == "CASE":
        return CASE(*translate(node.args, ctx))

    elif fn == "CAST":
        args = translate(node.args, ctx)
        if len(args) == 2 and isinstance(args[1], LIT) and isinstance(args[1].val, str):
            return FUN(S("CAST"), args[0], KW(S("AS"), OP(S(args[1].val))))
        else:
            return FUN(S("CAST"), *args)

    elif fn == "EXTRACT":
        args = translate(node.args, ctx)
        if len(args) == 2 and isinstance(args[0], LIT) and isinstance(args[0].val, str):
            return FUN(S("EXTRACT"), OP(S(args[0].val)), KW(S("FROM"), args[1]))
        else:
            return FUN(S("EXTRACT"), *args)

    elif fn in ("BETWEEN", "NOT BETWEEN"):
        if len(node.args) == 3:
            args = translate(node.args, ctx)
            return OP(S(fn), args[0], args[1], KW(S("AND"), over=args[2]))
        else:
            pass  # handled by the default logic

    elif fn in ("CURRENT_DATE", "CURRENT_TIMESTAMP"):
        if len(node.args) == 0:
            return OP(S(fn))
        else:
            pass  # handled by the default logic

    args = translate(node.args, ctx)
    if fn.isalpha():
        return FUN(S(fn), *args)
    else:
        return OP(S(fn), *args)


@translate_scalar.register
def _(node: Lit, ctx: TranslateContext) -> SQLClause:
    return LIT(node.val)


@translate_scalar.register
def _(node: Sort, ctx: TranslateContext) -> SQLClause:
    return SORT(value=node.value, nulls=node.nulls, over=translate(node.over, ctx))


@translate_scalar.register
def _(node: Var, ctx: TranslateContext) -> SQLClause:
    if node._name in ctx.vars_:
        return ctx.vars_[node._name]
    else:
        return VAR(node._name)


# -----------------------------------------------------------
# translating subquery nodes
# -----------------------------------------------------------


@singledispatch
def assemble(node: SQLNode, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    raise NotImplementedError(f"unhandled node of type: {type(node)}")


@assemble.register
def _(node: Box, ctx: TranslateContext) -> Assemblage:
    refs_p: list[SQLNode] = []
    for ref in node.refs:
        if isinstance(ref, HandleBound) and ref.handle == node.handle:
            refs_p.append(ref.over)
        else:
            refs_p.append(ref)
    base = assemble(node.over, refs_p, ctx)

    repl_p: NODE_TO_ALIAS = dict()
    for ref in node.refs:
        if isinstance(ref, HandleBound) and ref.handle == node.handle:
            repl_p[ref] = base.repl[ref.over]
        else:
            repl_p[ref] = base.repl[ref]
    return Assemblage(base.clause, base.cols, repl_p)


@assemble.register
def _(node: None, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    assert len(refs) == 0
    return Assemblage(None)


def aligned_columns(
    refs: list[SQLNode], repl: NODE_TO_ALIAS, args: list[SQLClause]
) -> bool:
    if len(refs) != len(args):
        return False
    for ref, arg in zip(refs, args):
        if not (isinstance(arg, (ID, AS)) and arg.name == repl[ref]):
            return False
    return True


@assemble.register
def _(node: Append, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    """no idea what is happening here"""
    assert node.over is not None
    base = assemble(node.over, ctx)
    branches = [(node.over, base)]
    for arg in node.args:
        branches.append((arg, assemble(arg, ctx)))

    dups: dict[SQLNode, SQLNode] = dict()
    seen: dict[Symbol, SQLNode] = dict()
    for ref in refs:
        name = base.repl[ref]
        if name in seen:
            other_ref = seen[name]
            if all(asmb.repl[ref] == asmb.repl[other_ref] for arg, asmb in branches):
                dups[ref] = seen[name]
        else:
            seen[name] = ref

    urefs = [ref for ref in refs if ref not in dups]
    repl, dummy_cols = refs_to_repl_n_cols(urefs)
    for ref, uref in dups.items():
        repl[ref] = repl[uref]

    cs: list[SQLClause] = []
    for arg, asmb in branches:
        if isinstance(asmb.clause, SELECT) and aligned_columns(
            urefs, repl, asmb.clause.args
        ):
            cs.append(asmb.clause)
            continue

        if not isinstance(asmb.clause, (SELECT, UNION)):
            alias = None
            tail = asmb.clause
        else:
            alias = ctx.allocate_alias(arg)
            tail = FROM(AS(over=asmb.get_complete_select(), name=alias))
        subs = asmb.get_make_subs(alias)
        cols: ALIAS_TO_CLAUSE = dict()
        for ref in urefs:
            name = repl[ref]
            cols[name] = subs[ref]
        clause = SELECT(*cols_to_select_args(cols), over=tail)
        cs.append(clause)

    clause = UNION(*cs[1:], all_=True, over=cs[0])
    return Assemblage(clause, cols=dummy_cols, repl=repl)


@assemble.register
def _(node: As, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    assert node.over is not None
    base = assemble(node.over, ctx)
    repl_p: NODE_TO_ALIAS = dict()
    for ref in refs:
        if isinstance(ref, NameBound):
            repl_p[ref] = base.repl[ref.over]
        else:
            repl_p[ref] = base.repl[ref]

    return Assemblage(base.clause, cols=base.cols, repl=repl_p)


@assemble.register
def _(node: Define, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    assert node.over is not None
    base = assemble(node.over, ctx)
    if not any(isinstance(ref, Get) and ref._name in node.label_map for ref in refs):
        return base

    if not isinstance(base.clause, (SELECT, UNION)):
        base_alias = None
        clause = base.clause
    else:
        base_alias = ctx.allocate_alias(node.over)
        clause = FROM(AS(name=base_alias, over=base.get_complete_select()))

    subs = base.get_make_subs(base_alias)
    translates: list[tuple[SQLNode, SQLClause]] = []
    tr_cache: ALIAS_TO_CLAUSE = dict()
    for ref in refs:
        if isinstance(ref, Get) and ref.over is None and ref._name in node.label_map:
            name = ref._name
            if name not in tr_cache:
                _define = node.args[node.label_map[name]]
                with ctx.substitute_vars_n_subs(subs=subs):
                    tr_cache[name] = translate(_define, ctx)
            col = tr_cache[name]
            translates.append((ref, col))
        else:
            translates.append((ref, subs[ref]))

    repl, cols = translates_to_repl_n_cols(translates)
    return Assemblage(clause, cols=cols, repl=repl)


@assemble.register
def _(node: FromNothing, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    return assemble(None, refs, ctx)


@assemble.register
def _(node: FromReference, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    cte_asmb = ctx.cte_map[node.over]
    asmb = cte_asmb.asmb.unwrap_repl()

    alias = ctx.allocate_alias(node.name)
    schema = None if cte_asmb.schema is None else ID(name=cte_asmb.schema)
    table = ID(name=cte_asmb.name, over=schema)
    clause = FROM(AS(name=alias, over=table))

    subs = asmb.get_make_subs(alias)
    translates = [(ref, subs[ref]) for ref in refs]
    repl, cols = translates_to_repl_n_cols(translates)
    return Assemblage(clause, cols=cols, repl=repl)


@assemble.register
def _(node: FromTable, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    seen: set[Symbol] = set()
    for ref in refs:
        assert (
            isinstance(ref, Get)
            and ref.over is None
            and ref._name in node.table.columns
        )
        seen.add(ref._name)
    alias = ctx.allocate_alias(node.table.name)

    schema = None if node.table.schema is None else ID(name=node.table.schema)
    table = ID(name=node.table.name, over=schema)
    clause = FROM(AS(name=alias, over=table))

    cols: ALIAS_TO_CLAUSE = dict()
    for col in node.table.columns:
        if col in seen:
            cols[col] = ID(name=col, over=ID(name=alias))

    repl: NODE_TO_ALIAS = dict()
    for ref in refs:
        if isinstance(ref, Get) and ref.over is None:
            repl[ref] = ref._name
    return Assemblage(clause, cols=cols, repl=repl)


@assemble.register
def _(node: FromValues, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    columns = node.source.columns
    column_set = set(columns)
    seen: set[Symbol] = set()
    for ref in refs:
        assert isinstance(ref, Get) and ref.over is None and ref._name in column_set
        seen.add(ref._name)

    rows: list[tuple]
    column_aliases: list[Symbol]
    # render the VALUES clause based on whether all/some/no columns are selected
    if len(seen) == len(columns):
        rows = node.source.data
        column_aliases = list(columns)
    elif len(seen) > 0:
        _indices = [i for i, c in enumerate(columns) if c in seen]
        rows = [tuple(row[i] for i in _indices) for row in node.source.data]
        column_aliases = [col for col in columns if col in seen]
    else:
        rows = [(None,) * len(node.source.data)]
        column_aliases = [S("_")]

    clause: SQLClause
    alias = ctx.allocate_alias(S("values"))
    cols: ALIAS_TO_CLAUSE = dict()
    if len(rows) == 0:
        clause = WHERE(LIT(False))
        for col in columns:
            if col in seen:
                cols[col] = LIT(None)
    elif not ctx.dialect.has_as_columns:
        # sqlite doesn't support column aliases for a VALUES clause
        col_prefix = ctx.dialect.values_column_prefix
        col_index = ctx.dialect.values_column_index
        assert col_prefix is not None
        clause = FROM(AS(alias, over=VALUES(rows)))
        parent = ID(name=alias)
        for col in columns:
            if col in seen:
                name = S(f"{col_prefix}{col_index}")
                cols[col] = ID(name=name, over=parent)
                col_index += 1
    else:
        clause = FROM(AS(alias, columns=column_aliases, over=VALUES(rows)))
        parent = ID(name=alias)
        for col in columns:
            if col in seen:
                cols[col] = ID(name=col, over=parent)

    repl: NODE_TO_ALIAS = dict()
    for ref in refs:
        if isinstance(ref, Get) and ref.over is None:
            repl[ref] = ref._name
    return Assemblage(clause, cols=cols, repl=repl)


@assemble.register
def _(node: Group, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    has_aggs = any(isinstance(ref, Agg) for ref in refs)
    if len(node.by) == 0 and not has_aggs:
        return assemble(None, refs, ctx)

    base = assemble(node.over, ctx)
    if base.clause is None or isinstance(base.clause, (FROM, JOIN, WHERE)):
        base_alias = None
        tail = base.clause
    else:
        assert node.over is not None
        base_alias = ctx.allocate_alias(node.over)
        tail = FROM(AS(name=base_alias, over=base.get_complete_select()))

    subs = base.get_make_subs(base_alias)
    with ctx.substitute_vars_n_subs(subs=subs):
        by = translate(node.by, ctx)
    translates: list[tuple[SQLNode, SQLClause]] = []
    for ref in refs:
        if isinstance(ref, Get) and ref.over is None:
            assert ref._name in node.label_map
            translates.append((ref, by[node.label_map[ref._name]]))
        elif isinstance(ref, Agg) and ref.over is None:
            with ctx.substitute_vars_n_subs(subs=subs):
                translates.append((ref, translate(ref, ctx)))

    repl, cols = translates_to_repl_n_cols(translates)
    assert len(cols) > 0
    if has_aggs:
        clause = GROUP(*by, over=tail)
    else:
        args = cols_to_select_args(cols)
        clause = SELECT(*args, distinct=True, over=tail)
        for name in cols:
            cols[name] = ID(name)
    return Assemblage(clause, cols=cols, repl=repl)


@assemble.register
def _(node: IntBind, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    vars_p = ctx.vars_.copy()
    for name, i in node.label_map.items():
        vars_p[name] = translate(node.args[i], ctx)
    with ctx.substitute_vars_n_subs(vars_=vars_p):
        return assemble(node.over, ctx)


@assemble.register
def _(node: IntIterate, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    with ctx.substitute_vars_n_subs(vars_=dict()):
        asmb_p = assemble(node.over, ctx)
    base = asmb_p.unwrap_repl()
    assert isinstance(base.clause, FROM)

    subs = base.get_make_subs(None)
    translates = []
    for ref in refs:
        translates.append((ref, subs[ref]))
    repl, cols = translates_to_repl_n_cols(translates)
    return Assemblage(base.clause, cols=cols, repl=repl)


@assemble.register
def _(node: IntJoin, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    assert node.over is not None
    left = assemble(node.over, ctx)
    if node.skip:
        return left
    if isinstance(left.clause, (FROM, JOIN)):
        left_alias = None
        tail = left.clause
    else:
        left_alias = ctx.allocate_alias(node.over)
        tail = FROM(AS(name=left_alias, over=left.get_complete_select()))

    subs = left.get_make_subs(left_alias)
    if len(node.lateral) > 0:
        with ctx.substitute_vars_n_subs(subs=subs):
            right = assemble(node.joinee, ctx)
    else:
        right = assemble(node.joinee, ctx)

    rclause = right.clause
    cond_1 = (
        isinstance(rclause, FROM)
        and isinstance(rclause.over, AS)
        and rclause.over.columns is None
        and isinstance(rclause.over.over, ID)
        and rclause.over.over.over is None
    )
    cond_2 = (
        isinstance(rclause, FROM)
        and isinstance(rclause.over, ID)
        and rclause.over.over is None
    )

    if cond_1 or cond_2:
        assert isinstance(rclause, FROM)  # make the typechecker happy
        joinee = rclause.over
        for ref, name in right.repl.items():
            subs[ref] = right.cols[name]
    else:
        right_alias = ctx.allocate_alias(node.joinee)
        joinee = AS(name=right_alias, over=right.get_complete_select())
        _cache: ALIAS_TO_CLAUSE = dict()
        for ref, name in right.repl.items():
            if name not in _cache:
                _cache[name] = ID(name=name, over=ID(right_alias))
            subs[ref] = _cache[name]

    with ctx.substitute_vars_n_subs(subs=subs):
        on = translate(node.on, ctx)
    assert joinee is not None
    clause = JOIN(
        joinee=joinee,
        on=on,
        left=node.left,
        right=node.right,
        lateral=len(node.lateral) > 0,
        over=tail,
    )
    translates = [(ref, subs[ref]) for ref in refs]
    repl, cols = translates_to_repl_n_cols(translates)

    return Assemblage(clause, cols=cols, repl=repl)


@assemble.register
def _(node: Knot, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    left = assemble(node.over, ctx)
    repl: NODE_TO_ALIAS = dict()
    seen: set[Symbol] = set()
    dups: set[SQLNode] = set()
    for ref in refs:
        assert isinstance(ref, NameBound)
        name = left.repl[ref.over]
        repl[ref] = name
        if name in seen:
            dups.add(ref.over)
        else:
            seen.add(name)

    temp_union = Assemblage(left.clause, cols=left.cols, repl=repl)
    union_alias = ctx.allocate_alias(node.name)
    ctx.cte_map[node.box] = CTEAssemblage(temp_union, name=union_alias)

    asmb = assemble(node.iterator, ctx)
    right = asmb.unwrap_repl()
    urefs = []
    for ref in refs:
        assert isinstance(ref, NameBound)
        if ref.over not in dups:
            urefs.append(ref.over)

    cs: list[SQLClause] = []
    assert node.over is not None
    _pairs: list[tuple[SQLNode, Assemblage]] = [
        (node.over, left),
        (node.iterator, right),
    ]
    for arg, asmb in _pairs:
        if isinstance(asmb.clause, SELECT) and aligned_columns(
            urefs, left.repl, asmb.clause.args
        ):
            cs.append(asmb.clause)
            continue
        elif not isinstance(asmb.clause, (SELECT, UNION)):
            alias = None
            tail = asmb.clause
        else:
            alias = ctx.allocate_alias(arg)
            tail = FROM(AS(name=alias, over=asmb.get_complete_select()))

        subs = asmb.get_make_subs(alias)
        cols: ALIAS_TO_CLAUSE = dict()
        for ref in urefs:
            name = left.repl[ref]
            cols[name] = subs[ref]
        cs.append(SELECT(*cols_to_select_args(cols), over=tail))

    union_clause = UNION(*cs[1:], all_=True, over=cs[0])
    cols: ALIAS_TO_CLAUSE = dict()
    for ref in urefs:
        name = left.repl[ref]
        cols[name] = ID(name)
    union = Assemblage(union_clause, cols=cols, repl=repl)
    ctx.cte_map[node.box] = CTEAssemblage(union, name=union_alias)
    ctx.recursive = True

    clause = FROM(over=ID(union_alias))
    subs = union.get_make_subs(union_alias)

    translates = [(ref, subs[ref]) for ref in refs]
    repl, cols = translates_to_repl_n_cols(translates)
    return Assemblage(clause, cols=cols, repl=repl)


@assemble.register
def _(node: Limit, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    assert node.over is not None
    base = assemble(node.over, ctx)
    if node.offset is None and node.limit is None:
        return base
    if base.clause is None or isinstance(
        base.clause, (FROM, JOIN, WHERE, GROUP, HAVING, ORDER)
    ):
        base_alias = None
        tail = base.clause
    else:
        base_alias = ctx.allocate_alias(node.over)
        tail = FROM(AS(name=base_alias, over=base.get_complete_select()))

    clause = LIMIT(node.limit, offset=node.offset, over=tail)
    subs = base.get_make_subs(base_alias)
    translates = [(ref, subs[ref]) for ref in refs]
    repl, cols = translates_to_repl_n_cols(translates)
    return Assemblage(clause, cols=cols, repl=repl)


@assemble.register
def _(node: Order, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    assert node.over is not None
    base = assemble(node.over, ctx)
    if len(node.by) == 0:
        return base

    if base.clause is None or isinstance(
        base.clause, (FROM, JOIN, WHERE, GROUP, HAVING)
    ):
        base_alias = None
        tail = base.clause
    else:
        base_alias = ctx.allocate_alias(node.over)
        tail = FROM(AS(name=base_alias, over=base.get_complete_select()))

    subs = base.get_make_subs(base_alias)
    with ctx.substitute_vars_n_subs(subs=subs):
        by = translate(node.by, ctx)
    clause = ORDER(*by, over=tail)
    translates = [(ref, subs[ref]) for ref in refs]
    repl, cols = translates_to_repl_n_cols(translates)
    return Assemblage(clause, cols=cols, repl=repl)


@assemble.register
def _(node: Partition, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    assert node.over is not None
    base = assemble(node.over, ctx)
    if not any(isinstance(ref, Agg) for ref in refs):
        return base

    if base.clause is None or isinstance(
        base.clause, (FROM, JOIN, WHERE, GROUP, HAVING)
    ):
        base_alias = None
        tail = base.clause
    else:
        base_alias = ctx.allocate_alias(node.over)
        tail = FROM(AS(name=base_alias, over=base.get_complete_select()))

    subs = base.get_make_subs(base_alias)
    clause = WINDOW(over=tail)
    with ctx.substitute_vars_n_subs(subs=subs):
        by = translate(node.by, ctx)
        order_by = translate(node.order_by, ctx)

        partition = PARTITION(*by, order_by=order_by, frame=node.frame)
        translates: list[tuple[SQLNode, SQLClause]] = []
        for ref in refs:
            if isinstance(ref, Agg) and ref.over is None:
                translates.append((ref, partition >> translate(ref, ctx)))
            else:
                translates.append((ref, subs[ref]))

    repl, cols = translates_to_repl_n_cols(translates)
    return Assemblage(clause, cols=cols, repl=repl)


@assemble.register
def _(node: Select, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    assert node.over is not None
    base = assemble(node.over, ctx)
    if not isinstance(base.clause, (SELECT, UNION)):
        base_alias = None
        tail = base.clause
    else:
        base_alias = ctx.allocate_alias(node.over)
        tail = FROM(AS(name=base_alias, over=base.get_complete_select()))

    subs = base.get_make_subs(base_alias)
    cols: ALIAS_TO_CLAUSE = dict()
    with ctx.substitute_vars_n_subs(subs=subs):
        for name, i in node.label_map.items():
            cols[name] = translate(node.args[i], ctx)
    clause = SELECT(*cols_to_select_args(cols), over=tail)

    cols = {name: ID(name) for name in cols}
    repl: NODE_TO_ALIAS = dict()
    for ref in refs:
        assert isinstance(ref, Get) and ref.over is None
        repl[ref] = ref._name
    return Assemblage(clause, cols=cols, repl=repl)


def merge_conditions(cond_1: SQLClause, cond_2: SQLClause) -> OP:
    is_and_cond_1 = isinstance(cond_1, OP) and str(cond_1.name) == "AND"
    is_and_cond_2 = isinstance(cond_2, OP) and str(cond_2.name) == "AND"

    # The typechecker can't tell cond_1 and cond_2 are actually `OP` objects, the
    # `isinstance` check doesn't carry through.
    if is_and_cond_1 and is_and_cond_2:
        return OP(S("AND"), *cond_1.args, *cond_2.args)  # type: ignore
    elif is_and_cond_1 and not is_and_cond_2:
        return OP(S("AND"), *cond_1.args, cond_2)  # type: ignore
    elif not is_and_cond_1 and is_and_cond_2:
        return OP(S("AND"), cond_1, *cond_2.args)  # type: ignore
    else:
        return OP(S("AND"), cond_1, cond_2)


@assemble.register
def _(node: Where, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    assert node.over is not None
    base = assemble(node.over, ctx)
    if (
        base.clause is None
        or isinstance(base.clause, (FROM, JOIN, WHERE, HAVING))
        or (isinstance(base.clause, GROUP) and len(base.clause.by) > 0)
    ):
        subs = base.get_make_subs(None)
        with ctx.substitute_vars_n_subs(subs=subs):
            cond = translate(node.condition, ctx)
        if isinstance(cond, LIT) and cond.val == True:
            return base
        if isinstance(base.clause, WHERE):
            cond = merge_conditions(base.clause.condition, cond)
            clause = WHERE(cond, over=base.clause.over)
        elif isinstance(base.clause, GROUP):
            clause = HAVING(cond, over=base.clause)
        elif isinstance(base.clause, HAVING):
            cond = merge_conditions(base.clause.condition, cond)
            clause = HAVING(cond, over=base.clause.over)
        else:
            clause = WHERE(cond, over=base.clause)
    else:
        base_alias = ctx.allocate_alias(node.over)
        tail = FROM(AS(name=base_alias, over=base.get_complete_select()))
        subs = base.get_make_subs(base_alias)
        with ctx.substitute_vars_n_subs(subs=subs):
            cond = translate(node.condition, ctx)
        if isinstance(cond, LIT) and cond.val == True:
            return base
        clause = WHERE(cond, over=tail)

    translates = [(ref, subs[ref]) for ref in refs]
    repl, cols = translates_to_repl_n_cols(translates)
    return Assemblage(clause, cols=cols, repl=repl)


@assemble.register
def _(node: With, refs: list[SQLNode], ctx: TranslateContext) -> Assemblage:
    for arg in node.args:
        asmb = assemble(arg, ctx)
        alias = ctx.allocate_alias(arg)
        ctx.cte_map[arg] = CTEAssemblage(
            asmb, name=alias, materialized=node.materialized
        )
    return assemble(node.over, ctx)
